require 'rails_helper'
require [Hermes::Engine.root, %w[lib extras string]].flatten.join('/')

RSpec.describe Hermes::Score, type: :model do
  let!(:plus) do
    create(
      :threshold,
      type: 'PLUS',
      thresholds_map: {
        'Paris et communes limitrophes' => {
          'thresholds_list' => [23_721, 35_452, 46_473, 55_486, 66_017, 74_286],
          'another_person' => 8278
        }, 'IDF hors Paris et Communes limitrophes' => {
          'thresholds_list' => [23_721, 35_452, 42_616, 51_046, 60_429, 68_001],
          'another_person' => 7577
        }, 'Autres régions' => {
          'thresholds_list' => [20_623, 27_540, 33_119, 39_982, 47_035, 53_008],
          'another_person' => 5912
        }
      }
    )
  end
  let!(:plai) do
    create(
      :threshold,
      type: 'PLA I',
      thresholds_map: {
        'Paris et communes limitrophes' => {
          'thresholds_list' => [13_050, 21_272, 27_883, 30_521, 36_307, 40_859],
          'another_person' => 4552
        }, 'IDF hors Paris et Communes limitrophes' => {
          'thresholds_list' => [13_050, 21_272, 25_569, 28_075, 33_238, 37_401],
          'another_person' => 4166
        }, 'Autres régions' => {
          'thresholds_list' => [11_342, 16_525, 19_872, 22_111, 25_870, 29_155],
          'another_person' => 3252
        }
      }
    )
  end
  context 'with basic setup' do
    let!(:applicant_1) do
      create(
        :applicant,
        monthly_incomes: {
          "Salaire": 1000
        }
      )
    end
    let!(:housing_request_1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant_id: applicant_1.id,
        rejected_in_cal_nb: 1,
        rejected_after_cal_nb: 0,
        priority_points_indexes: [30, 145]
      )
    end

    let!(:entity_1) do
      create(
        :entity
      )
    end

    let!(:applicant_2) do
      create(
        :applicant,
        birth_date: Date.parse((Date.current - 10.years).to_s),
        entity: entity_1,
        monthly_incomes: {
          "Salaire": 2000
        }
      )
    end
    let!(:housing_request_2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant_id: applicant_2.id,
        rejected_in_cal_nb: 0,
        rejected_after_cal_nb: 1,
        entity_validated: true,
        priority_points_indexes: [30, 145]
      )
    end

    let!(:housing_offer_1) do
      create(
        :housing_offer,
        rent_amount: 500
      )
    end

    it 'has a valid factory' do
      expect(
        FactoryBot.build(
          :score,
          applicant: applicant_1,
          housing_offer: housing_offer_1
        )
      ).to be_valid
    end

    describe 'Base relationship and creation' do
      it 'should has rightly set relationships' do
        expect(applicant_1.housing_requests.length).to be > 0
        expect(applicant_2.housing_requests.length).to be > 0
      end

      it 'should has set applicants to have at least one active request' do
        expect(applicant_1.housing_requests.actives.first.is_active).to be true
        expect(applicant_2.housing_requests.actives.first.is_active).to be true
        expect(applicant_1.first_active_request).to_not be_nil
        expect(applicant_2.first_active_request).to_not be_nil
      end
    end

    let!(:score_1) do
      create(
        :score,
        is_active: true,
        applicant_id: applicant_1.id,
        housing_offer_id: housing_offer_1.id
      )
    end

    let!(:score_2) do
      create(
        :score,
        is_active: true,
        applicant_id: applicant_2.id,
        housing_offer_id: housing_offer_1.id,
        epci: 24
      )
    end

    describe 'Test scoring process' do
      before :each do
        score_1.process
        score_2.process
      end

      it 'should update data with right types' do
        expect([score_1.profile_score, score_2.profile_score]).to(
          all(be_a(Integer))
        )
        expect([score_1.profile_epci_score, score_2.profile_epci_score]).to(
          all(be_a(Integer))
        )
      end

      it 'should eventualy has set the right profile_score' do
        expect(score_1.any_applicant_is_under_30).to be false
        expect(score_1.entity_points).to eq 0
        expect(score_1.already_rejected_in_cal).to be true
        expect(score_1.already_declined_cal).to be false
        expect(score_1.calculate_history_points).to eq 35
        expect(score_1.calculate_effort_rate_points).to eq 35
        expect(score_1.profile_score.to_i + score_1.epci.to_i).to be(
          score_1.profile_epci_score
        )

        expect(score_2.any_applicant_is_under_30).to be true
        expect(score_2.entity_points).to be > 0
        expect(score_2.already_rejected_in_cal).to be false
        expect(score_2.already_declined_cal).to be true
        expect(score_2.calculate_history_points).to eq(-35)
        expect(score_2.calculate_effort_rate_points).to be < 35
        expect(score_2.profile_score.to_i + score_2.epci.to_i).to be(
          score_2.profile_epci_score
        )
      end
    end
  end

  context 'with precise setup' do
    # ENTITIES
    let!(:entity1) { create(:entity, siret: '123456789') }
    let!(:entity2) { create(:entity, siret: '2398479930') }
    let!(:entity3) { create(:entity, siret: '1234556778') }
    # FAMILY 1
    let!(:coapp1) do
      create(
        :applicant,
        birth_date: '1985-12-14',
        first_name: 'Jacqueline',
        last_name: 'Durand',
        family_situation: 'Marié (e)',
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1250
        },
        entity: entity2
      )
    end
    let!(:account1) { create(:user, first_name: 'Jean', last_name: 'Dupont', email: 'jean.dupont@test.test') }
    let!(:app1) do
      create(
        :applicant,
        user_id: account1.id.to_s,
        nud: '1223455',
        sne_profile_creation_date: '2016-04-03',
        family_situation: 'Marié (e)',
        birth_date: '1995-03-12',
        year_minus_one: '2018',
        amount_year_minus_one: 29_400,
        year_minus_two: '2017',
        amount_year_minus_two: 29_400,
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1200
        },
        entity: entity1
      )
    end
    let!(:request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: app1,
        seeked_departments: %w[
          75
        ],
        seeked_districts: %w[
          92002
          92014
          92071
        ],
        coapplicants: [
          { rel_type: 'Marié (e)', id: coapp1.id }
        ],
        dependents: [
          {
            rel_type: 'enfant à charge',
            birth_date: '2017-03-23',
            first_name: 'Paul',
            last_name: 'Dupont'
          },
          {
            rel_type: 'enfant à charge',
            birth_date: '2019-08-04',
            first_name: 'Cerise',
            last_name: 'Dupont'
          }
        ]
      )
    end
    # FAMILY 2
    let!(:coapp2) do
      create(
        :applicant,
        birth_date: '1985-12-14',
        first_name: 'Jeanne',
        last_name: 'Alphonso',
        family_situation: 'Pacsé (e)',
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1050
        },
        entity: entity2
      )
    end
    let!(:account2) { create(:user, first_name: 'Eric', last_name: 'Alphonso', email: 'eric.alphonso@test.test') }
    let!(:app2) do
      create(
        :applicant,
        user_id: account2.id.to_s,
        nud: '445559980',
        sne_profile_creation_date: '2017-06-13',
        family_situation: 'Pacsé (e)',
        birth_date: '1995-03-12',
        year_minus_one: '2018',
        amount_year_minus_one: 28_200,
        year_minus_two: '2017',
        amount_year_minus_two: 28_200,
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1300
        },
        entity: entity3
      )
    end
    let!(:request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: app2,
        seeked_departments: %w[
          75
        ],
        seeked_districts: %w[
          92002
          92014
          92071
        ],
        coapplicants: [
          { rel_type: 'Pacsé (e)', id: coapp2.id }
        ],
        dependents: [
          {
            rel_type: 'enfant à charge',
            birth_date: '2017-03-23',
            first_name: 'Grégoire',
            last_name: 'Alphonso'
          }
        ]
      )
    end
    # FAMILY 3
    let!(:coapp3) do
      create(
        :applicant,
        birth_date: '1985-12-14',
        first_name: 'Paulette',
        last_name: 'Rimeri',
        family_situation: 'Pacsé (e)',
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1375
        },
        entity: entity2
      )
    end
    let!(:account3) { create(:user, first_name: 'François', last_name: 'D\'arville', email: 'francois.darville@test.test') }
    let!(:app3) do
      create(
        :applicant,
        user_id: account3.id.to_s,
        nud: '678993039',
        sne_profile_creation_date: '2017-08-23',
        family_situation: 'Pacsé (e)',
        birth_date: '1995-03-12',
        year_minus_one: '2018',
        amount_year_minus_one: 33_900,
        year_minus_two: '2017',
        amount_year_minus_two: 33_900,
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1450
        },
        entity: entity3
      )
    end
    let!(:request3) do
      create(
        :housing_request,
        is_active: true,
        mutation: true,
        main_applicant: app3,
        seeked_departments: %w[
          75
        ],
        seeked_districts: %w[
          92002
          92014
          92071
        ],
        coapplicants: [
          { rel_type: 'Pacsé (e)', id: coapp3.id }
        ],
        dependents: [
          {
            rel_type: 'enfant à charge',
            birth_date: '2017-03-23',
            first_name: 'George',
            last_name: 'Rimeri',
            monthly_incomes: {
              'Salaire ou revenu d\'activité': 2825
            }
          }
        ]
      )
    end
    # OFFERS
    let!(:offer1) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 57,
        postal_code: '92002',
        department: '92',
        threshold_type: 'PLUS',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 640,
        rental_charges: 50,
      )
    end
    let!(:offer2) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 53,
        postal_code: '92071',
        department: '92',
        threshold_type: 'PLUS',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 750,
        rental_charges: 60
      )
    end
    let!(:offer3) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 70,
        postal_code: '92002',
        department: '92',
        threshold_type: 'PLA I',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 560,
        rental_charges: 90
      )
    end
    let!(:offer4) do
      create(
        :housing_offer,
        typology: 'T2',
        surface: 80,
        postal_code: '92014',
        department: '92',
        threshold_type: 'PLUS',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: '900',
        rental_charges: '110'
      )
    end
    let!(:offer5) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 80,
        postal_code: '92012',
        department: '92',
        threshold_type: 'PLUS',
        threshold_subtype: 'Paris et communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 1000,
        rental_charges: 110
      )
    end
    let!(:offer6) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 85,
        postal_code: '92012',
        department: '92',
        threshold_type: 'PLA I',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 700,
        rental_charges: 110
      )
    end
    let!(:offer7) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 90,
        postal_code: '92071',
        department: '92',
        threshold_type: 'PLA I',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 500,
        rental_charges: 110
      )
    end
    let!(:offer8) do
      create(
        :housing_offer,
        typology: 'T4',
        surface: 95,
        postal_code: '92071',
        department: '92',
        threshold_type: 'PLUS',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 1450,
        rental_charges: 110
      )
    end
    let!(:offer9) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 45,
        department: '75',
        threshold_type: 'PLA I',
        threshold_subtype: 'Paris et communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 760,
        rental_charges: 50
      )
    end
    let!(:offer10) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 45,
        department: '75',
        threshold_type: 'PLUS',
        threshold_subtype: 'Paris et communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 760,
        rental_charges: 50
      )
    end
    let!(:offer11) do
      create(
        :housing_offer,
        typology: 'T2',
        surface: 45,
        department: '75',
        threshold_type: 'PLUS',
        threshold_subtype: 'Paris et communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 760,
        rental_charges: 50
      )
    end

    before :each do
      perform_enqueued_jobs do
        enqueued_jobs.each { |job| job[:job].perform_now(*job[:args]) }
      end
    end

    describe 'check eligibility and scores for all families' do
      describe 'eligibility and score for family1' do
        it 'should has valid eligiblity' do
          r_app1 = app1.reload
          where = r_app1.eligibility_query.json_to_h
          eligible_offers = Hermes::HousingOffer.where(where)
          eligible_ids = eligible_offers.pluck(:id).map(&:to_s)
          expect(request1.reload.category_type).to be 4
          expect(request1.reload.reference_tax_revenues).to(
            be > plai.thresholds_map['IDF hors Paris et Communes limitrophes']['thresholds_list'][3]
          )          
          expect(eligible_offers.count).to be 4
          [offer1, offer2, offer9, offer10].each do |offer|
            expect(eligible_ids).to include(offer.id.to_s)
          end
        end
      end

      describe 'eligibility and score for family2' do
        let(:r_app2) { app2.reload }

        it 'should has valid eligiblity' do
          where = r_app2.eligibility_query.json_to_h
          eligible_offers = Hermes::HousingOffer.where(where.merge(offer_status: 'LOGEMENT PUBLIE'))
          eligible_ids = eligible_offers.pluck(:id).map(&:to_s)
          expect(request2.category_type).to be 3
          expect(eligible_offers.count).to be 1
          [offer1].each do |offer|
            expect(eligible_ids).to include(offer.id.to_s)
          end
        end

        it 'has generated valid scores' do
          r2_scores = Hermes::Score.where(applicant: r_app2)
          expect(r2_scores.count).to be 1
          expect(r2_scores.pluck(:mutation)).not_to include(false)
        end
      end

      describe 'eligibility and score for family3' do
        let(:r_app3) { app3.reload }

        it 'should has valid eligiblity' do
          where = r_app3.eligibility_query.json_to_h
          eligible_offers = Hermes::HousingOffer.where(where)
          eligible_ids = eligible_offers.pluck(:id).map(&:to_s)
          expect(request3.category_type).to be 3
          expect(eligible_offers.count).to be 4
          [offer1, offer2, offer10, offer11].each do |offer|
            expect(eligible_ids).to include(offer.id.to_s)
          end
        end

        it 'has generated valid scores' do
          r3_scores = Hermes::Score.where(applicant: r_app3)
          expect(r3_scores.count).to be 4
          expect(r3_scores.pluck(:mutation)).not_to include(false)
        end
      end
    end
  end

  context 'for eligibility_count' do
    let!(:entity1) { create(:entity, siret: '123456789') }
    let!(:entity2) { create(:entity, siret: '2398479930') }
    let!(:coapp1) do
      create(
        :applicant,
        birth_date: '1985-12-14',
        first_name: 'Jacqueline',
        last_name: 'Durand',
        family_situation: 'Marié (e)',
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1250
        },
        entity: entity2
      )
    end
    let!(:account1) { create(:user, first_name: 'Jean', last_name: 'Dupont', email: 'jean.dupont2@test.test') }
    let!(:app1) do
      create(
        :applicant,
        user_id: account1.id.to_s,
        nud: '1223455',
        sne_profile_creation_date: '2016-04-03',
        family_situation: 'Marié (e)',
        birth_date: '1995-03-12',
        year_minus_one: '2018',
        amount_year_minus_one: 29_400,
        year_minus_two: '2017',
        amount_year_minus_two: 29_400,
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1200
        },
        entity: entity1
      )
    end
    let!(:request1) do
      create(
        :housing_request,
        is_active: true,
        entity_validated: true,
        main_applicant: app1,
        mutation: true,
        seeked_departments: %w[
          75
        ],
        seeked_districts: %w[
          92002
          92014
          92071
        ],
        coapplicants: [
          { rel_type: 'Marié (e)', id: coapp1.id }
        ],
        dependents: [
          {
            rel_type: 'enfant à charge',
            birth_date: '2017-03-23',
            first_name: 'Paul',
            last_name: 'Dupont'
          },
          {
            rel_type: 'enfant à charge',
            birth_date: '2019-08-04',
            first_name: 'Cerise',
            last_name: 'Dupont'
          }
        ]
      )
    end

    let!(:account2) { create(:user, first_name: 'Jean', last_name: 'Dupont', email: 'jean.dupont3@test.test') }
    let!(:app2) do
      create(
        :applicant,
        user_id: account1.id.to_s,
        nud: '1223455',
        sne_profile_creation_date: '2016-04-03',
        family_situation: 'Marié (e)',
        birth_date: '1995-03-12',
        year_minus_one: '2018',
        amount_year_minus_one: 29_400,
        year_minus_two: '2017',
        amount_year_minus_two: 29_400,
        monthly_incomes: {
          'Salaire ou revenu d\'activité': 1200
        },
        entity: entity2
      )
    end
    let!(:request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: app2,
        seeked_departments: %w[
          75
        ],
        seeked_districts: %w[
          92002
          92014
          92071
        ],
        coapplicants: [
          { rel_type: 'Marié (e)', id: coapp1.id }
        ],
        dependents: [
          {
            rel_type: 'enfant à charge',
            birth_date: '2017-03-23',
            first_name: 'Paul',
            last_name: 'Dupont'
          },
          {
            rel_type: 'enfant à charge',
            birth_date: '2019-08-04',
            first_name: 'Cerise',
            last_name: 'Dupont'
          }
        ]
      )
    end

    let!(:offer1) do
      create(
        :housing_offer,
        typology: 'T3',
        surface: 57,
        district: '92002',
        department: '92',
        threshold_type: 'PLUS',
        threshold_subtype: 'IDF hors Paris et Communes limitrophes',
        offer_status: 'LOGEMENT PUBLIE',
        min_effort_rate: 0.333,
        rent_amount: 640,
        rental_charges: 50
      )
    end

    before :each do
      perform_enqueued_jobs do
        enqueued_jobs.each { |job| job[:job].perform_now(*job[:args]) }
      end
    end

    describe 'eligibility quantities are valide for offer1' do
      let(:r_offer1) { offer1.reload }

      it 'has valid quantity of eligibility' do
        expect(r_offer1.eligible_requests_nb).to be 2
      end

      it 'has valid eligible details' do
        expect(r_offer1.eligible_requests_nb_detailed['mutation']).to be 1
        expect(r_offer1.eligible_requests_nb_detailed['entities'][app1.entity_id.to_s]).to be 1
        expect(r_offer1.eligible_requests_nb_detailed['entities'][app2.entity_id.to_s]).to_not be_present
      end
    end
  end
  def time_rand from = Time.parse("1997-01-01"), to = Time.now
    Time.at(from + rand * (to.to_f - from.to_f))
  end

  def rand_mail
    rand_mail_ext = -> { %w[com net org fr].sample }
    "#{Faker::Name.name_with_middle.downcase.gsub(' ', '.')}@test.#{rand_mail_ext.call}"
  end

  def rand_typology
    "T#{rand(1..12)}"
  end

  def rand_subtype
    [
      'Paris et communes limitrophes',
      'IDF hors Paris et Communes limitrophes',
      'Autres régions'
    ].sample
  end

  context 'with large quantity of requests' do
    let!(:threshold1) do
      create(
        :threshold,
        type: 'PLA LM',
        thresholds_map: {
          'Paris et communes limitrophes': {
            'thresholds_list': [
              13_050,
              21_272,
              27_883,
              30_521,
              36_307,
              40_859
            ],
            'another_person': 4552
          },
          'IDF hors Paris et Communes limitrophes': {
            'thresholds_list': [
              13_050,
              21_272,
              25_569,
              28_075,
              33_238,
              37_401
            ],
            'another_person': 4166
          },
          'Autres régions': {
            'thresholds_list': [
              11_342,
              16_525,
              19_872,
              22_111,
              25_870,
              29_155
            ],
            'another_person': 3252
          }
        }
      )
    end

    let!(:applicants_list) {
      puts 'in applicants_list creation'
      100.times.map do
        user = create(
          :user,
          email: rand_mail
        )
        applicant = create(
          :applicant,
          user_id: user.id,
          amount_year_minus_two: rand(7000..10_000),
          monthly_incomes: {
            'SAL' => rand(760.00..3500.00)
          }
        )
        request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          seeked_types: ['T2'],
          seeked_departments: ['34']
        )
        nud = create(
          :nud,
          department: '34',
          creation_date: time_rand,
          housing_request: request
        )
        {
          user: user,
          applicant: applicant,
          request: request,
          nud: nud
        }
      end
    }

    let!(:offers_list) do
      puts 'in offers_list creation'
      1.times.map do
        create(
          :housing_offer,
          typology: 'T2',
          department: '34',
          offer_status: 'LOGEMENT PUBLIE',
          rent_amount: rand(100.00..890.00),
          threshold_type: 'PLA LM',
          threshold_subtype: rand_subtype
        )
      end
    end

    describe 'with all scores generated' do
      before :each do
        puts "in before with #{applicants_list.length * offers_list.length } objects"
        t = Time.now
        expect(Hermes::HousingRequest.count).to be applicants_list.length
        expect(Hermes::HousingOffer.count).to be offers_list.length
        perform_enqueued_jobs do
          enqueued_jobs.each { |job| job[:job].perform_now(*job[:args]) }
        end
        puts "performed queue in #{Time.now - t}"
      end

      it 'should has valid quantity of created scores and should be correctly ordinated' do
        expect(Hermes::Score.count).to be(applicants_list.length * offers_list.length)
        expect(Hermes::Score.only(:sne_date).pluck(:sne_date)).to all(be_a(DateTime).or(be_a(Time)))
        offer_ids = Hermes::Score.only(:housing_offer_id).pluck(:housing_offer_id).uniq.compact
        offer_ids.each do |offer_id|
          offer_scores = Hermes::Score.where(housing_offer_id: offer_id).order(profile_epci_score: -1, sne_date: 1)
          offer_scores.each_cons(2) do |a, b|
            # puts "now : [#{a.profile_epci_score}]#{a.sne_date} " \
            #   "next : [#{b.profile_epci_score}]#{b.sne_date}"
            expect(b.eligible_position).to be >= a.eligible_position
          end
        end
      end
    end
  end
end
